package controller

import (
	"electricity_bill_calc/cache"
	"electricity_bill_calc/exceptions"
	"electricity_bill_calc/global"
	"electricity_bill_calc/model"
	"electricity_bill_calc/response"
	"electricity_bill_calc/security"
	"electricity_bill_calc/service"
	"electricity_bill_calc/tools"
	"fmt"
	"net/http"
	"strconv"

	"github.com/gofiber/fiber/v2"
	"github.com/shopspring/decimal"
)

func InitializeUserController(router *fiber.App) {
	router.Delete("/password/:uid", security.OPSAuthorize, invalidUserPassword)
	router.Delete("/login", security.MustAuthenticated, logout)
	router.Put("/password", resetUserPassword)
	router.Get("/accounts", security.OPSAuthorize, listPagedUser)
	router.Post("/login", login)
	router.Put("/account/enabled/state", security.OPSAuthorize, switchUserEnabling)
	router.Post("/account", security.OPSAuthorize, createOPSAndManagementAccount)
	router.Get("/account/:uid", security.MustAuthenticated, getUserDetail)
	router.Post("/enterprise", security.OPSAuthorize, createEnterpriseAccount)
	router.Put("/account/:uid", security.OPSAuthorize, modifyAccountDetail)
	router.Get("/enterprise/quick/search", security.OPSAuthorize, quickSearchEnterprise)
	router.Get("/expiration", security.EnterpriseAuthorize, fetchExpiration)
}

type _LoginFormData struct {
	Username string `json:"uname"`
	Password string `json:"upass"`
	Type     int8   `json:"type"`
}

func login(c *fiber.Ctx) error {
	result := response.NewResult(c)
	loginData := new(_LoginFormData)
	if err := c.BodyParser(loginData); err != nil {
		return result.Error(http.StatusInternalServerError, "表单解析失败。")
	}
	var (
		session *model.Session
		err     error
	)
	if loginData.Type == model.USER_TYPE_ENT {
		session, err = service.UserService.ProcessEnterpriseUserLogin(loginData.Username, loginData.Password)
	} else {
		session, err = service.UserService.ProcessManagementUserLogin(loginData.Username, loginData.Password)
	}
	if err != nil {
		if authError, ok := err.(*exceptions.AuthenticationError); ok {
			if authError.NeedReset {
				return result.LoginNeedReset()
			}
			return result.Error(int(authError.Code), authError.Message)
		} else {
			return result.Error(http.StatusInternalServerError, err.Error())
		}
	}
	return result.LoginSuccess(session)
}

func logout(c *fiber.Ctx) error {
	result := response.NewResult(c)
	session := c.Locals("session")
	if session == nil {
		return result.Success("用户会话已结束。")
	}
	_, err := cache.ClearSession(session.(*model.Session).Token)
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	return result.Success("用户已成功登出系统。")
}

func invalidUserPassword(c *fiber.Ctx) error {
	result := response.NewResult(c)
	targetUserId := c.Params("uid")
	verifyCode, err := service.UserService.InvalidUserPassword(targetUserId)
	if _, ok := err.(exceptions.NotFoundError); ok {
		return result.NotFound("未找到指定用户。")
	}
	if _, ok := err.(exceptions.UnsuccessfulOperationError); ok {
		return result.NotAccept("未能成功更新用户的密码。")
	}
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	return result.Json(http.StatusAccepted, "用户密码已经失效", fiber.Map{"verify": verifyCode})
}

type _ResetPasswordFormData struct {
	VerifyCode  string `json:"verifyCode"`
	Username    string `json:"uname"`
	NewPassword string `json:"newPass"`
}

func resetUserPassword(c *fiber.Ctx) error {
	result := response.NewResult(c)
	resetForm := new(_ResetPasswordFormData)
	if err := c.BodyParser(resetForm); err != nil {
		return result.UnableToParse("无法解析提交的数据。")
	}
	verified, err := service.UserService.VerifyUserPassword(resetForm.Username, resetForm.VerifyCode)
	if _, ok := err.(exceptions.NotFoundError); ok {
		return result.NotFound("指定的用户不存在。")
	}
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	if !verified {
		return result.Error(http.StatusUnauthorized, "验证码不正确。")
	}
	completed, err := service.UserService.ResetUserPassword(resetForm.Username, resetForm.NewPassword)
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	if completed {
		return result.Updated("用户凭据已更新。")
	}
	return result.NotAccept("用户凭据未能成功更新。")
}

func listPagedUser(c *fiber.Ctx) error {
	result := response.NewResult(c)
	requestPage, err := strconv.Atoi(c.Query("page", "1"))
	if err != nil {
		return result.NotAccept("查询参数[page]格式不正确。")
	}
	requestKeyword := c.Query("keyword")
	requestUserType, err := strconv.Atoi(c.Query("type", "-1"))
	if err != nil {
		return result.NotAccept("查询参数[type]格式不正确。")
	}
	var requestUserStat *bool
	state, err := strconv.ParseBool(c.Query("state"))
	if err != nil {
		requestUserStat = nil
	} else {
		requestUserStat = &state
	}
	users, total, err := service.UserService.ListUserDetail(requestKeyword, requestUserType, requestUserStat, requestPage)
	if err != nil {
		return result.NotFound(err.Error())
	}
	return result.Json(
		http.StatusOK,
		"已取得符合条件的用户集合。",
		response.NewPagedResponse(requestPage, total).ToMap(),
		fiber.Map{"accounts": users},
	)
}

type _UserStateChangeFormData struct {
	UserID  string `json:"uid" form:"uid"`
	Enabled bool   `json:"enabled" form:"enabled"`
}

func switchUserEnabling(c *fiber.Ctx) error {
	result := response.NewResult(c)
	switchForm := new(_UserStateChangeFormData)
	if err := c.BodyParser(switchForm); err != nil {
		return result.UnableToParse("无法解析提交的数据。")
	}
	err := service.UserService.SwitchUserState(switchForm.UserID, switchForm.Enabled)
	if err != nil {
		if nfErr, ok := err.(*exceptions.NotFoundError); ok {
			return result.NotFound(nfErr.Message)
		} else {
			return result.Error(http.StatusInternalServerError, err.Error())
		}
	}
	return result.Updated("用户状态已经更新。")
}

type _OPSAccountCreationFormData struct {
	Username string  `json:"username" form:"username"`
	Name     string  `json:"name" form:"name"`
	Contact  *string `json:"contact" form:"contact"`
	Phone    *string `json:"phone" form:"phone"`
	Type     int     `json:"type" form:"type"`
}

func createOPSAndManagementAccount(c *fiber.Ctx) error {
	result := response.NewResult(c)
	creationForm := new(_OPSAccountCreationFormData)
	if err := c.BodyParser(creationForm); err != nil {
		return result.UnableToParse("无法解析提交的数据。")
	}
	exists, err := service.UserService.IsUsernameExists(creationForm.Username)
	if exists {
		return result.Conflict("指定的用户名已经被使用了。")
	}
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	newUser := new(model.User)
	newUser.Username = creationForm.Username
	newUser.Type = int8(creationForm.Type)
	newUser.Enabled = true
	newUserDetail := new(model.UserDetail)
	newUserDetail.Name = &creationForm.Name
	newUserDetail.Contact = creationForm.Contact
	newUserDetail.Phone = creationForm.Phone
	newUserDetail.UnitServiceFee = decimal.Zero
	newUserDetail.ServiceExpiration, _ = model.ParseDate("2099-12-31")
	verifyCode, err := service.UserService.CreateUser(newUser, newUserDetail)
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	cache.AbolishRelation("user")
	return result.Json(http.StatusCreated, "用户已经成功创建。", fiber.Map{"verify": verifyCode})
}

func getUserDetail(c *fiber.Ctx) error {
	result := response.NewResult(c)
	targetUserId := c.Params("uid")
	exists, err := service.UserService.IsUserExists(targetUserId)
	if !exists {
		return result.NotFound("指定的用户不存在。")
	}
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	userDetail, err := service.UserService.FetchUserDetail(targetUserId)
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	return result.Json(http.StatusOK, "用户详细信息已获取到。", fiber.Map{"user": userDetail})
}

type _EnterpriseCreationFormData struct {
	Username       string  `json:"username" form:"username"`
	Name           string  `json:"name" form:"name"`
	Region         *string `json:"region" form:"region"`
	Address        *string `json:"address" form:"address"`
	Contact        *string `json:"contact" form:"contact"`
	Phone          *string `json:"phone" form:"phone"`
	UnitServiceFee *string `json:"unitServiceFee" form:"unitServiceFee"`
}

func createEnterpriseAccount(c *fiber.Ctx) error {
	result := response.NewResult(c)
	creationForm := new(_EnterpriseCreationFormData)
	if err := c.BodyParser(creationForm); err != nil {
		return result.UnableToParse("无法解析提交的数据。")
	}
	exists, err := service.UserService.IsUsernameExists(creationForm.Username)
	if exists {
		return result.Conflict("指定的用户名已经被使用了。")
	}
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	newUser := new(model.User)
	newUser.Username = creationForm.Username
	newUser.Type = model.USER_TYPE_ENT
	newUser.Enabled = true
	newUserDetail := new(model.UserDetail)
	newUserDetail.Name = &creationForm.Name
	newUserDetail.Contact = creationForm.Contact
	newUserDetail.Phone = creationForm.Phone
	newUserDetail.UnitServiceFee, err = decimal.NewFromString(*creationForm.UnitServiceFee)
	if err != nil {
		return result.BadRequest("用户月服务费无法解析。")
	}
	newUserDetail.ServiceExpiration = model.NewEmptyDate()

	verifyCode, err := service.UserService.CreateUser(newUser, newUserDetail)
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	cache.AbolishRelation("user")
	return result.Json(http.StatusCreated, "用户已经成功创建。", fiber.Map{"verify": verifyCode})
}

type _AccountModificationFormData struct {
	Name           string  `json:"name" form:"name"`
	Region         *string `json:"region" form:"region"`
	Address        *string `json:"address" form:"address"`
	Contact        *string `json:"contact" form:"contact"`
	Phone          *string `json:"phone" form:"phone"`
	UnitServiceFee *string `json:"unitServiceFee" form:"unitServiceFee"`
}

func modifyAccountDetail(c *fiber.Ctx) error {
	result := response.NewResult(c)
	targetUserId := c.Params("uid")
	modForm := new(_AccountModificationFormData)
	if err := c.BodyParser(modForm); err != nil {
		return result.UnableToParse("无法解析提交的数据。")
	}
	exists, err := service.UserService.IsUserExists(targetUserId)
	if !exists {
		return result.NotFound("指定的用户不存在。")
	}
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	newUserInfo := new(model.UserDetail)
	newUserInfo.Id = targetUserId
	newUserInfo.Name = &modForm.Name
	if len(modForm.Name) > 0 {
		abbr := tools.PinyinAbbr(modForm.Name)
		newUserInfo.Abbr = &abbr
	}
	newUserInfo.Region = modForm.Region
	newUserInfo.Address = modForm.Address
	newUserInfo.Contact = modForm.Contact
	newUserInfo.Phone = modForm.Phone
	newUserInfo.UnitServiceFee, err = decimal.NewFromString(*modForm.UnitServiceFee)
	if err != nil {
		return result.BadRequest("用户月服务费无法解析。")
	}
	_, err = global.DB.NewUpdate().Model(newUserInfo).
		WherePK().
		Column("name", "abbr", "region", "address", "contact", "phone", "unit_service_fee").
		Exec(c.Context())
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	cache.AbolishRelation(fmt.Sprintf("user:%s", targetUserId))
	return result.Updated("指定用户的信息已经更新。")
}

func quickSearchEnterprise(c *fiber.Ctx) error {
	result := response.NewResult(c)
	keyword := c.Query("keyword")
	searchResult, err := service.UserService.SearchLimitUsers(keyword, 6)
	if err != nil {
		return result.Error(http.StatusInternalServerError, err.Error())
	}
	return result.Json(http.StatusOK, "已查询到存在符合条件的企业", fiber.Map{"users": searchResult})
}

func fetchExpiration(c *fiber.Ctx) error {
	result := response.NewResult(c)
	session, err := _retreiveSession(c)
	if err != nil {
		return result.Unauthorized(err.Error())
	}
	user, err := service.UserService.FetchUserDetail(session.Uid)
	if err != nil {
		return result.NotFound(err.Error())
	}
	return result.Json(
		http.StatusOK,
		"已经取得用户的服务期限信息",
		fiber.Map{"expiration": user.ServiceExpiration.Format("2006-01-02")},
	)
}
